#include <stdio.h>
#include <stdlib.h>

#include "agrippa-cipher.h"

const unsigned cell_length = 3;

const unsigned e_2 = 11;
const unsigned p_q = 4097; // 17*241 appears to have been chosen due to its proximity to 2^12 for 3-char packing

typedef unsigned (*bitmapper)(unsigned);

// various maps of 0..7 -> 0..7
unsigned un_do_too_it(unsigned x) { return ((13 - x) * 5) % 8; }
unsigned un_do_it(unsigned x) { return (((7 - x) * 3) + 5) % 8; }

// byte scrambler helper
unsigned bytemapper(bitmapper f, unsigned x) {
    byte ret = 0;
    for (unsigned bit = 0; bit < 8; bit++)
      if (x & (1<<bit)) {
	unsigned newbit = f(bit);
	ret |= (1 << newbit);
      }

    return ret;
}

// two byte scramblers
byte un_waymute_it(byte x) { return bytemapper(un_do_too_it,x); }
byte un_permute_it(byte x) { return bytemapper(un_do_it,x); }

// A 12-bit scrambler (cf. RSA, despite the name "des")
unsigned un_make_des(unsigned x) {
  unsigned ret = 1;
  for (int i = 0; i < e_2; i++)
    ret = (ret * x ) % p_q;

  return ret;
}

// Map old Mac's carriage return to newline
void output(byte c) {
  if (c == '\r') c = '\n';
  fputc(c,stdout);
}

int main() {
  for (int cell = 0 ; cell < agrippa_cipherlen / cell_length ; cell++) {
    unsigned char c1, c2, c3;

    c1 = agrippa_cipher[cell*3+0];
    c2 = agrippa_cipher[cell*3+1];
    c3 = agrippa_cipher[cell*3+2];

    // scramble just the 2nd character
    c2 = un_waymute_it(c2);

    // "un-happenin"

    // combine 3 characters (24 bits) into two 12-bits numbers, which are then scrambled
    unsigned this_num = un_make_des( ((c2 % 16) * 256) | c1);
    unsigned that_num = un_make_des( (c3 * 16) | (c2 / 16) );
    
    // chop the 24 bits back into 3 characters
    unsigned temp1 = this_num % 256;
    unsigned temp2 = ((that_num % 16) * 16) | (this_num / 256);
    unsigned temp3 = that_num / 16;

    // scramble each character, and output
    output(un_permute_it(temp1));
    output(un_permute_it(temp2));
    output(un_permute_it(temp3));
  } 

  printf("\n");
  return 0;
}
